#!/usr/bin/env python3
import sys
import syslog
import subprocess
import xml.etree.ElementTree as ET

try:
    import collectd
except ImportError:
    # Handle the case where collectd is not available
    collectd = None

# This is the name of the plugin which will be set it the collectd.conf too.
PLUGIN_NAME = 'gpu'


DEVICE_SPEC = [
    {1: 'product_name'},
    {2: 'driver_version'},
    {3: 'vbios_version'},
]

UTILIZATION_PARAMETERS = [
    {1: 'gpu_util'},
    {2: 'memory_util'},
    {3: 'fan_speed'},
]

TEMPERATURE = [
    {1: 'gpu_temp'}
]


MEMORY_USAGE = [
    {1: 'total'},
    {2: 'reserved'},
    {3: 'used'},
    {4: 'free'}
]


CLOCKS = [
    {1: 'graphics_clock'},
    {2: 'sm_clock'},
    {3: 'mem_clock'},
    {4: 'video_clock'}
]


EVENT_REASONS = [
    {1: 'clocks_event_reason_gpu_idle'},
    {2: 'clocks_event_reason_hw_thermal_slowdown'},
    {3: 'clocks_event_reason_sw_power_cap'},
    {4: 'clocks_event_reason_applications_clocks_setting'},
    {5: 'clocks_event_reason_hw_power_brake_slowdown'},
    {6: 'clocks_event_reason_sw_thermal_slowdown'},
    {7: 'clocks_event_reason_sync_boost'},
    {8: 'clocks_event_reason_display_clocks_setting'},
    {9: 'clocks_event_reason_hw_slowdown'}
]

PCIE_GEN = [
    {1: 'current_link_gen'},
    {2: 'device_current_link_gen'},
]

ENCODED_STATS = [
    {1: 'session_count'},
    {2: 'average_fps'},
    {3: 'average_latency'}
]

POWER_DRAW = [
    {1: 'power_draw'}
]


def log(param):
    """
    Log messages to either collectd or stdout depending on how it was called.

    :param param: the message
    :return:  None
    """

    if __name__ != '__main__':
        collectd.info("%s: %s" % (PLUGIN_NAME, param))
    else:
        sys.stderr.write("%s\n" % param)


def get_gpu_info(metrics, xml_data):
    lines = []
    lines.append('-' * 70)
    heading = f"{'GPU Information':^35}"
    lines.append(heading)

    gpu_ids = [gpu.get("id") for gpu in xml_data.findall(".//gpu")]

    for gpu_id in gpu_ids:
        gpu_info = {'GPU id': gpu_id}

        for param_dict in metrics:
            for param_num, param_name in param_dict.items():
                element = xml_data.find(f".//gpu[@id='{gpu_id}']/{param_name}")
                value = element.text.strip() if element is not None else "N/A"
                gpu_info[param_name.capitalize().replace('_', ' ')] = value

        # Check for driver_version outside the loop
        driver_version_element = xml_data.find(f".//gpu[@id='{gpu_id}']/../driver_version")
        gpu_info['Driver version'] = driver_version_element.text.strip() if driver_version_element is not None else "N/A"

        max_key_length = max(len(key) for key in gpu_info.keys())
        format_string = f"{{:<{max_key_length}}} | {{:<20}}"

        lines.append('-' * 70)
        for key, value in gpu_info.items():
            lines.append(format_string.format(key, value))
        lines.append('-' * 70)

    print("\n".join(lines))


def dispatch_values(metrics, gpu_id, root):
    for param_dict in metrics:
        for param_num, param_name in param_dict.items():
            for element in root.iter(param_name):
                try:
                    value = int(element.text.strip().split(' ')[0])
                    syslog.syslog(syslog.LOG_ERR, f"Parameter {param_num}: {param_name} - Value: {value}")

                    collectd.Values(plugin=PLUGIN_NAME,
                                    type_instance=param_name,
                                    plugin_instance=gpu_id,
                                    type="gauge",
                                    values=[value]).dispatch()
                except AttributeError:
                    syslog.syslog(syslog.LOG_ERR,
                                  f"AttributeError: 'text' attribute not found for {param_name} in gpu_id {gpu_id}")
                    # Handle the AttributeError (e.g., set a default value or log an error message)
                except ValueError:
                    syslog.syslog(syslog.LOG_ERR,
                                  f"ValueError: Unable to convert {param_name} value to int in gpu_id {gpu_id}")
                    # Handle the ValueError (e.g., set a default value or log an error message)
                except Exception as e:
                    syslog.syslog(syslog.LOG_ERR, f"An unexpected error occurred: {e}")
                    # Handle other unexpected errors


def dispatch_values_with_path(xml_path, metrics, gpu_id, root, instance_prefix=''):
    path = root.find(xml_path)

    if path is not None:
        for param_dict in metrics:
            for param_num, param_name in param_dict.items():
                param_element = path.find(param_name)

                if param_element is not None:
                    try:
                        value = float(param_element.text.strip().split(' ')[0])
                        syslog.syslog(syslog.LOG_ERR, f"Parameter {param_num}: {param_name} - Value: {value}")

                        collectd.Values(plugin=PLUGIN_NAME,
                                        type_instance=f"{instance_prefix}{param_name}",
                                        plugin_instance=gpu_id,
                                        type="gauge",
                                        values=[value]).dispatch()
                    except AttributeError:
                        syslog.syslog(syslog.LOG_ERR,
                                      f"AttributeError: 'text' attribute not found for {param_name} in gpu_id {gpu_id}")
                        # Handle the AttributeError (e.g., set a default value or log an error message)
                    except ValueError:
                        syslog.syslog(syslog.LOG_ERR,
                                      f"ValueError: Unable to convert {param_name} value to float in gpu_id {gpu_id}")
                        # Handle the ValueError (e.g., set a default value or log an error message)
                    except Exception as e:
                        syslog.syslog(syslog.LOG_ERR, f"An unexpected error occurred: {e}")
                        # Handle other unexpected errors
                else:
                    syslog.syslog(syslog.LOG_ERR, f"Parameter {param_name} not found in xml_path for gpu_id {gpu_id}")
                    # You can add further handling logic if needed


def dispatch_value_boolean(xml_path, metrics, gpu_id, root):
    clock_event_reasons = root.find(xml_path)
    syslog.syslog(syslog.LOG_ERR, f"There are clock event reasons {clock_event_reasons}")

    if clock_event_reasons is not None:
        for param_dict in metrics:
            for param_num, param_name in param_dict.items():
                # Iterate over elements under <fb_memory_usage> and extract values
                param_element = clock_event_reasons.find(param_name)

                if param_element is not None:
                    try:
                        value = str(param_element.text.strip())
                        if value == "Active":
                            value = 0
                        elif value == 'Not Active':
                            value = 1
                        else:
                            # Handle the case when the value is neither "Active" nor "Not Active"
                            syslog.syslog(syslog.LOG_ERR, f"Unexpected value '{value}' for the clock event reason "
                                                          f"{param_name} for gpu_id {gpu_id}")
                            # You can add further handling logic if needed

                        syslog.syslog(syslog.LOG_ERR, f"The value for the clock event reason is {value}")

                        collectd.Values(plugin=PLUGIN_NAME,
                                        type_instance=param_name,
                                        plugin_instance=gpu_id,
                                        type="gauge",
                                        values=[value]).dispatch()
                    except AttributeError:
                        syslog.syslog(syslog.LOG_ERR,
                                      f"AttributeError: 'text' attribute not found for {param_name} in gpu_id {gpu_id}")
                        # Handle the AttributeError (e.g., set a default value or log an error message)
                    except Exception as e:
                        syslog.syslog(syslog.LOG_ERR, f"An unexpected error occurred: {e}")
                        # Handle other unexpected errors
                else:
                    syslog.syslog(syslog.LOG_ERR,
                                  f"Parameter {param_name} not found in clock_event_reasons for gpu_id {gpu_id}")
                    # You can add further handling logic if needed


def dispatch_values_splitting(xml_path, metrics, gpu_id, root, str_to_remove=' '):
    path = root.find(xml_path)

    if path is not None:
        for param_dict in metrics:
            for param_num, param_name in param_dict.items():
                param_element = path.find(param_name)

                if param_element is not None:
                    try:
                        value = float(param_element.text.strip().split(str_to_remove)[0])
                        syslog.syslog(syslog.LOG_ERR, f"Parameter {param_num}: {param_name} - Value: {value}")
                        collectd.Values(plugin=PLUGIN_NAME,
                                        type_instance=param_name,
                                        plugin_instance=gpu_id,
                                        type="gauge",
                                        values=[value]).dispatch()
                    except AttributeError:
                        syslog.syslog(syslog.LOG_ERR,
                                      f"AttributeError: 'text' attribute not found for {param_name} in gpu_id {gpu_id}")
                        # Handle the AttributeError (e.g., set a default value or log an error message)
                    except ValueError:
                        syslog.syslog(syslog.LOG_ERR,
                                      f"ValueError: Unable to convert {param_name} value to float in gpu_id {gpu_id}")
                        # Handle the ValueError (e.g., set a default value or log an error message)
                else:
                    syslog.syslog(syslog.LOG_ERR, f"Parameter {param_name} not found in xml_path for gpu_id {gpu_id}")
                    # You can add further handling logic if needed


def metric_show():
    try:
        # Get the GPU xml file
        nvidia_smi_output = subprocess.check_output(["nvidia-smi", "-q", "-x"]).decode("utf-8")
        root = ET.fromstring(nvidia_smi_output)
        # get GPUs ids
        gpu_ids = [gpu.get("id") for gpu in root.findall(".//gpu")]

        for gpu_id in gpu_ids:
            dispatch_values_with_path(f".//gpu[@id='{gpu_id}']/fb_memory_usage", MEMORY_USAGE, gpu_id, root, 'memory_')
            dispatch_values_with_path(f".//gpu[@id='{gpu_id}']/pci/pci_gpu_link_info/pcie_gen", PCIE_GEN, gpu_id, root)
            dispatch_values_with_path(f".//gpu[@id='{gpu_id}']/clocks", CLOCKS, gpu_id, root)
            dispatch_values_with_path(f".//gpu[@id='{gpu_id}']/gpu_power_readings", CLOCKS, gpu_id, root)
            dispatch_values_with_path(f".//gpu[@id='{gpu_id}']/encoder_stats", ENCODED_STATS, gpu_id, root)
            dispatch_values(UTILIZATION_PARAMETERS, gpu_id, root)
            dispatch_values(TEMPERATURE, gpu_id, root)
            dispatch_values_with_path(f".//gpu[@id='{gpu_id}']/gpu_power_readings", POWER_DRAW, gpu_id, root)
            dispatch_value_boolean(f".//gpu[@id='{gpu_id}']/clocks_event_reasons", EVENT_REASONS, gpu_id, root)

    except subprocess.CalledProcessError as e:
        print(f"Error: {e}")


def config(conf):
    """
    This method has been registered as the config callback and is used to parse options from
    given config.  Note that you cannot receive the whole config files this way, only Module blocks
    inside the Python configuration block. Additionally you will only receive blocks where your
    callback identifier matches your plugin.

    In this case Frequency is a float value that will modify the frequency of the sine wave. This
    in conjunction with the polling interval can give you as smooth or blocky a curve as you want.

    :param conf: a Config object
    :return: None
    """

    for kv in conf.children:
        if kv.key == 'Frequency':
            global FREQUENCY
            FREQUENCY = float(kv.values[0])


def init():
    """
    This method has been registered as the init callback; this gives the plugin a way to do startup
    actions. We'll just log a message.

    :return: None
    """

    log("Plugin %s initializing..." % PLUGIN_NAME)


if __name__ != "__main__":
    # when running inside plugin register each callback
    collectd.register_read(metric_show)
    collectd.register_config(config)
    collectd.register_init(init)
else:
    nvidia_smi_output = subprocess.check_output(["nvidia-smi", "-q", "-x"]).decode("utf-8")
    root = ET.fromstring(nvidia_smi_output)
    # gpu_id = get_gpu_id()
    get_gpu_info(DEVICE_SPEC, root)
